#!/usr/bin/python
#
# Program to insert/replace commands in snap & proc files to enable FILA10G
# & Flexbuff operations.
#
# v1 2014-04-29 S. Casey
# Initial version
#
# v2 2014-05-28 S. Casey
# Include decimation for FILA10G v3.2_280314
#
# v3 2014-06-03 S. Casey
# Parse bitmask to send correct number of bits to flexbuff in mode=vdif:bits:vdif_framsize
# and include net_protocol, mtu.
#
# v4 2014-06-13 S. Casey
# Move FILA start command to end of a def block where other fila commands have been issued
# instead of starting/stopping for each scan
#
# v5 2014-06-18 S. Casey
# Add vdif_frame command. Set VDIF frame size to 8000
#
# v6 2014-07-01 S. Casey
# New FILA firmware includes VDIF sequence numbers
#
# v7 2014-10-16 S. Casey
# Change mode= command to new VDIF_8000-1024-X-Y... format for jive5ab-2.5.1
#
# v8 2014-10-24 S. Casey
# Add support for disk2fileFB via ssh & autoftpFB
#
# v9 2014-10-31 S. Casey
# Change write block size 256MB -> 1024MB
#
# Rename fila5cflex.py
# v000 2014-11-11 S. Casey
# Add support for multiple recorders & Mk5C
#
# v001 2014-11-11 S. Casey
# Add option to record with net2file on flexbuff
#
# v002 2015-02-26 S. Casey
# Add FS 9.11.72 support = sample rate instead of decimation on mk5b mode command
#
# v003 2015-03-09 S. Casey
# Add support for alternate FILA10G script e.g. for internal or external
#
# v004 2015-04-01 S. Casey
# Add support for mk6 mode on flexbuff-like devices & set_disks= to specify recording locations


import sys

def usage():
  print "Usage: %s expt_name mk5b/vdif [flex/flex_NN/5c] [flex/flex_NN/5c] [alt_fila] [set_disks=location1:location2:....]\n" % sys.argv[0]
  print "    Where expt_name is the name of the experiment to be modified"
  print "    Second parameter specifies VDIF or Mark5B format"
  print "    Third & fourth optional parameters of 'flex' 'flex_NN', '5c' or 'mk6' indicates that extra commands"
  print "    for flexbuff and/or Mark5C and/or Mk6 should be inserted for parallel recordings."
  print "    If recorder type is mk6, disks= must be specified,"
  print "    'flex_NN' records on flexbuff as a single file for each scan. All scans are recorded in"
  print "    /mnt/diskNN/scan_name where scan_name is derived from input snap filename"
  print "    'alt_fila' specifies to use alternate script to communcate with FILA10G "
  print "    The string following disks= must be a valid string to pass to jive5ab's set_disks command and must be"
  print "    present if recorder type is Mk6, and can optionally be present for regular flexbuff recordings"
  sys.exit(0)

def Bits(i):    #Python 2.5 doesn't have a bin() :(
  i = i - ((i >> 1) & 0x55555555)
  i = (i & 0x33333333) + ((i >> 2) & 0x33333333)
  return (((i + (i >> 4) & 0xF0F0F0F) * 0x1010101) & 0xffffffff) >> 24

#How to send commands to FILA10G
#
#fila_cmd = dbbc=fila10g=
fila_cmd = "sy=cmd2fila10g.py "

#altfila_cmd = dbbc=fila10g=
alt_fila_cmd = "sy=cmd2fila10g_int.py "

#How to send commands to Flexbuff if piggybacking
flex_cmd = "sy=cmd2flexbuff.py "
mk5c_cmd = "sy=cmd2mk5c.py "

#Default clock rate (MHz) to assume if no clock_set
default_clock=32

#Location of snap & proc files:
snap = "/usr2/sched/"
proc = "/usr2/proc/"

#Station ID
station_id = "on"
if len(sys.argv) < 3:
  usage()

# jive5ab mode=vdif:tracks:framsize
# framesize excludes 32 byte vdif header!
# odd quirk - mtu bust be > IP packet size. mtu = IP packet size doesn't work.

if sys.argv[2] == "vdif":
  format="vdif"
  framesize=8000 # 8000 payload + 32 byte header
  mtu = framesize + 32 + 8 + 28 + 2 # 32 byte vdif header + 8 bytes sequence + 28 bytes UDP/IP header + 2 mtu quirk
  ##temp fix for missing vdif sequence numbers"
#  mtu = framesize + 32 + 0 + 28 + 1 # 32 byte vdif header + 0 bytes sequence + 28 bytes UDP/IP header + 1 mtu quirk
  protocol="udps"
  mk5c_pkt = "36:0:" + str(framesize + 32) + ":0:0"
elif sys.argv[2] == "mk5b":
  format="mk5b"
  framesize=5008 - 32 # Reduce framesize by vdif header size 32 bytes
  protocol="udps"
  mtu = framesize + 32 + 8 + 28 + 4 # 32 bytes vdif header + 8 bytes sequence + 28 bytes UDP/IP header + 4 mtu quirk
  mk5c_pkt = "36:0:" + str(framesize + 32) + ":0:0"
else:
  usage()

piggyback_flex = False
piggyback_5c = False
mk6 = 0
flex_disk=""
set_disks=""

arg_n = 3
n_rec = 0

for i, arg in enumerate(sys.argv[3:]):
  if arg.lower() == "flex":
    if n_rec >= 2:
      print "Too many recorders (max 2)!)"
      usage()
    piggyback_flex=True
    n_rec += 1
  elif arg.lower()[:5] == "flex_":
    if n_rec >= 2:
      print "Too many recorders (max 2)!)"
      usage()
    try:
      a=int(arg[5:])
      if a > 40:
        print "Warning:", a, " is probably not a valid flexbuff disk number"
      flex_disk=arg[5:]
    except ValueError:
      print "Error:", arg, "does not end with a valid number"
      usage()
    piggyback_flex=True
    n_rec += 1
  elif arg.lower() == "5c":
    if n_rec >= 2:
      print "Too many recorders (max 2)!)"
      usage()
    piggyback_5c=True
    n_rec += 1
  elif arg.lower() == "mk6":
    if n_rec >= 2:
      print "Too many recorders (max 2)!)"
      usage()
    piggyback_flex=True
    mk6 = 1
    n_rec += 1
  elif arg.lower() == "alt_fila":
    fila_cmd = alt_fila_cmd
  elif arg[:9].lower() == "set_disks":
    disk_str = "".join(sys.argv[i + 3:])
    if disk_str[9] == "=":
      set_disks = disk_str[10:]
      break    # disks=..... must be last parameter
    else:
      print "Invalid format specified for set_disk list"
      usage()
  else:
    print "Unrecognised parameter found:", arg
    usage()

if mk6 == 1 and len(set_disks) == 0:
  print "Error: set_disks= must be provided when recorder type is mk6"
  usage()

piggyback = piggyback_flex or piggyback_5c
flex = piggyback_flex or (not piggyback)
mk5c = piggyback_5c
flexS = len(flex_disk) > 0
expt = sys.argv[1] + station_id
flex_path = "/mnt/disk" + flex_disk + "/" + expt + "/"

print "Flex piggyback:", piggyback_flex
print "Mk5C piggyback:", piggyback_5c
print "Mk6 mode:", mk6
print "Flex enabled:", flex
print "Flex single disk mode:", flexS
print "Mk5C enabled:", mk5c
if flexS:
  print "Recording on flexbuff to " + flex_path
  print "Please ensure that directory exists"

try:
  for ext in ".prc", ".snp":
    if ext == ".prc":
      dir=proc
    else:
      dir=snap
    fname=dir + "/" + expt + ext
    print "Opening " + fname + " for reading..."
    fIn = open(fname)

    fnameOut=fname + ".out"
    print "Opening " + fnameOut + " for writing..."
    fOut = open(fnameOut, "w")

    lines = fIn.readlines()
    define = -1
    enddef = -1
    seen_mode = -1
    seen_clock = 0
    start_fila = 0

    decimation = 1
    clock_rate = default_clock

    for ln in range(0, len(lines)):
      if lines[ln][:6] == "define":
        define = ln   # mark start of define block
        for dn in range(ln + 1, len(lines)):
          if lines[dn][:6] == "define":
            print "Uh oh, found a define within a define block on line number", dn, "! Giving up."
            sys.exit(1)
          if lines[dn][:6] == "enddef":
            enddef = dn
            break
        else:        # Hmmm didn't find an enddef
          print "Uh oh, found a define on line", dn, "that had no enddef before end of file! Giving up."
          sys.exit(1)
        fOut.write(lines[ln])

      elif lines[ln][:6] == "enddef":
        define = -1
        enddef = -1
        seen_mode = -1
        seen_clock = 0
        decimation = 1
        clock_rate = default_clock
        filter_bandwidth = default_clock 

        if start_fila:
          fOut.write(fila_cmd + "start " + format + "\n")
          start_fila = 0
        fOut.write(lines[ln])

      elif lines[ln][:10] == "scan_name=":
        sn=lines[ln][10:].split(",")
        scan_name=sn[1]+"_"+sn[2]+"_"+sn[0]
        fOut.write(lines[ln])

      elif lines[ln][:10] == "disk2file=":
        sn=lines[ln][10:].split(",")
        if piggyback:
          fOut.write(lines[ln])
        if sn[0] != "abort":
         fOut.write("sy=disk2fileFB.py " + lines[ln][10:].rstrip() + "&\n")

      elif lines[ln][:14] == "disk_record=on":
        fOut.write(lines[ln])
        if piggyback_flex:
          if flexS:
            fOut.write(flex_cmd + "net2file=open:" + flex_path + scan_name + ",n ;\n")
          else:
            fOut.write(flex_cmd + "record = on : " + scan_name + " ;\n")
        if piggyback_5c:
          fOut.write(mk5c_cmd + "record = on : " + scan_name + " ;\n")
#        fOut.write(fila_cmd + "start " + format + "\n")

      elif lines[ln][:15] == "disk_record=off":
#        fOut.write(fila_cmd +"stop\n")
        fOut.write(lines[ln])
        if piggyback_flex:
          if flexS:
            fOut.write(flex_cmd + "net2file = close\n")
          else:
            fOut.write(flex_cmd + "record = off\n")
        if piggyback_5c:
          fOut.write(mk5c_cmd +"record = off"+"\n")

      elif lines[ln][:9] == "mk5=mode=":
        print lines[ln]
        params=lines[ln][9:].split(':')
        fOut.write(lines[ln])
        if piggyback_flex:
          fOut.write(flex_cmd + lines[ln][4:])
#        fOut.write(fila_cmd + "vsi_bitmask " + params[1]+"\n")

      elif lines[ln][:10] == "mk5b_mode=":
        print lines[ln]
        start_fila=1   # Insert fila start command at end of this def block 
        params=lines[ln][10:].split(',')
        print params[0], params[1], params[2]
#        fOut.write(lines[ln])  ##jive5ab doesn't like mode=bitmask decimation
        bits=Bits( int(params[1],16) )
        if Bits(bits) != 1: #Non power of 2 bitmask
          print "Error: bitmask", lines[ln][5], "is not a power of 2"
          sys.exit(1)
        flexcmd="mk5="
        if piggyback:
          fOut.write(lines[ln]) ## only add in mode=bitmask for regular mk5 if doing piggyback
          flexcmd=flex_cmd
        if(int(params[1],16) & 0xAAAAAAAA):
          bits_p_chan=2
        else:
          bits_p_chan=1
        print "Detected", bits_p_chan, "bits per channel"
#        fOut.write(flexcmd + "mode=vdif:" + str(bits) + ":" + str(framesize) + "\n")     # Meh "vdif" for both vdif & mk5b :) 
        if flex: # Commands needed only for flexbuff ops
          if flexS:
            fOut.write(flexcmd + "\"net_protocol=" + protocol + ":32000000:1000000:16; mtu=" + str(mtu) + "\"\n") # 32MB socket buffer, 1MB write block size
          else:
            fOut.write(flexcmd + "\"net_protocol=" + protocol + ":32000000:512000000:8; mtu=" + str(mtu) + "; record = mk6:" + mk6 + "\"\n") # 32MB socket buffer, 512MB write block size
            fOut.write(flexcmd + "\"set_disks=" + disks_list + "\"\n")
        if mk5c:
          fOut.write(mk5c_cmd + "packet=" + mk5c_pkt + "\n")
        fOut.write(fila_cmd + "vsi_bitmask " + params[1] + "\n")
        if format == "vdif":
          fOut.write(fila_cmd + "vdif_frame " + str(bits_p_chan) + " " + str(bits / bits_p_chan) + " " + str(framesize) + "\n")

        # Mode also contains decimation value and/or clock , which is sent to FILA with clock value...
        if len(params) == 3:
          decimation = int(params[2])
        elif len(params) == 4:
          try:
            decimation = int(params[2])
          except:
            decimation = 1
          filter_bandwidth = int(float(params[3]))

        # Look to see if we have a clock in this section
        if define != -1:      # We are in a define section
          end = enddef
        else:
          end = len(lines)

        # Only search for clock if we didn't alredy see a clock_set yet
        if seen_clock == 0 :
          for dn in range(ln, end):
            if lines[dn][:14] == "mk5=clock_set=":
              # Grab clock value here
              print "Found clock_set on line", dn
              params=lines[ln][14:].split(':')
              clock_rate = int(params[0])
        fOut.write(fila_cmd + "vsi_samplerate " + str(clock_rate * 1000000) + " " + str(decimation * (clock_rate / (filter_bandwidth))) + "\n")
        if flex:
          if flexS:
            fOut.write(flexcmd + "mode=none\n")
          else:
            fOut.write(flexcmd + "mode=VDIF_" + str(framesize) + "-" + str(bits * clock_rate / decimation) + "-" + str(bits / bits_p_chan) + "-" + str(bits_p_chan) + "\n")     # Meh "vdif" for both vdif & mk5b :)

      elif lines[ln][:14] == "mk5=clock_set=":
        start_fila=1   # Insert fila start command at end of this def block 
        params=lines[ln][14:].split(':')
        fOut.write(lines[ln])
        if piggyback:
          fOut.write(flex_cmd + lines[ln][4:])
        clock_rate = int(params[0])
        print "Found clock_set on line", clock_rate
#        fOut.write(fila_cmd + "vsi_samplerate " + params[0]+"000000 " + str(decimation))
#        fOut.write(flexcmd + "mode=VDIF_" + str(frameisze) + "-" + str(bits * ( int(params[0] / decimation) + "-" + str(bits / bits_p_chan) + "-" + str(bits_p_chan) + "\n")     # Meh "vdif" for both vdif & mk5b :)
        seen_clock = 1
      else:
        fOut.write(lines[ln])

except IOError:
  print "Could not open", fname
